import uuid
from queue import Queue
from queue import Empty


class SimpleQueue:

    def __init__(self, ack=False):
        self._queue = Queue()
        self._dict = {}
        self._ack = ack

    def put(self, data):
        _qid = str(uuid.uuid1())
        self._queue.put((data, _qid))
        self._dict[_qid] = data
        return _qid

    def put_batch(self, data_list):
        qid_list = [self.put(_item) for _item in data_list]
        return qid_list

    def get(self, block=False, timeout=None):
        try:
            _data, _qid = self._queue.get(block=block, timeout=timeout)
            if not self._ack:
                del self._dict[_qid]
            return _data, _qid
        except Empty:
            return None, None

    def ack(self, qid):
        if not self._ack:
            return True
        if qid is None:
            return False
        del self._dict[qid]
        return True

    def queue_size(self):
        return self._queue.qsize()

    def dict_size(self):
        return len(self._dict)


class UniqueKeyAgentClient:

    def __init__(self, register_manager, topic_name):
        self._topic_name = topic_name
        self._register_manager = register_manager
        self._register_manager.bind(topic_name)

    def send(self, key, data):
        qid = self._register_manager.put(self._topic_name, key, data)
        return qid

    def receive(self):
        data, qid = self._register_manager.get(self._topic_name)
        return data, qid

    def commit(self, qid):
        result = self._register_manager.ack(self._topic_name, qid)
        return result

    def send_batch(self, data_list: list):
        qid_list = self._register_manager.put_batch(self._topic_name, data_list)
        return qid_list

    def receive_batch(self, max_num):
        datas = []
        for i in range(max_num):
            data, qid = self._register_manager.get(self._topic_name)
            if data is None:
                break
            datas.append((data, qid))
        return datas

    # def receive_batch(self, num):
    #     datas = [self._register_manager.get(self._topic_name) for i in range(num)]
    #     return datas

    def commit_batch(self, qid_list):
        results = [self._register_manager.ack(self._topic_name, qid) for qid in qid_list]
        return results



from loguru import logger
from .simple_queue import SimpleQueue


class SimpleTopicAgent:

    _topics = {}

    _instance = None

    def __new__(cls, *args, **kwargs):
        # If no instance of class already exits
        if cls._instance is None:
            cls._instance = object.__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self, *args, **kwargs):
        if self._initialized:
            return
        # set the attribute to `True` to not initialize again
        self._initialized = True

    def bind(self, topic_name):
        if topic_name in self._topics:
            logger.debug(f'topic name {topic_name} already exist')
            return False
        self._topics[topic_name] = SimpleQueue()
        return True

    def put(self, topic_name, data):
        q = self._topics[topic_name]
        qid = q.put(data)
        logger.debug(f'topic: {topic_name}, put data: {data}, qid: {qid}')
        return qid

    def put_batch(self, topic_name, data_list):
        q = self._topics[topic_name]
        qid_list = q.put_batch(data_list)
        logger.debug(f'topic: {topic_name}, put data_list: {data_list}, qid_list: {qid_list}')
        return qid_list

    def get(self, topic_name):
        q = self._topics[topic_name]
        data, qid = q.get()
        if data is None:
            logger.debug(f'topic: {topic_name}, get data: {data}, qid: {qid}')
        else:
            logger.debug(f'topic: {topic_name}, get data: {data}, qid: {qid}')
        return data, qid

    def ack(self, topic_name, qid):
        q = self._topics[topic_name]
        result = q.ack(qid)
        logger.debug(f'topic: {topic_name}, ack qid: {qid}')
        return result

    def info(self, ):
        desp = {name: f'queue size = {q.queue_size()}, dict size = {q.dict_size()}' for name, q in self._topics.items()}
        return desp



import sys
from multiprocessing.managers import BaseManager
from .brokers.simple_topic_agent import SimpleTopicAgent
from .brokers.unique_key_topic_agent import UniqueKeyTopicAgent
from flask import Flask
from flask import jsonify
from .common.protocol import DEFAULT_SIMPLE_AGENT_NAME
from .common.protocol import DEFAULT_UNIQUE_KEY_AGENT_NAME
from loguru import logger
import threading

logger.remove()
# logger.add(sys.stderr, level='DEBUG')

app = Flask(__name__)

unique_key_agent = UniqueKeyTopicAgent()
simple_agent = SimpleTopicAgent()

@app.route('/')
def index():
    resp = {
        'UniqueKeyAgent': unique_key_agent.info(),
        'SimpleAgent': simple_agent.info(),
    }
    return jsonify(resp)


class QueueManager(BaseManager):
    pass


QueueManager.register(DEFAULT_SIMPLE_AGENT_NAME, SimpleTopicAgent)
QueueManager.register(DEFAULT_UNIQUE_KEY_AGENT_NAME, UniqueKeyTopicAgent)


class SmqServer():


    def __init__(self, api_port, queue_port, auth_key, log_level='INFO'):
        self._api_port = api_port
        self._queue_port = queue_port
        self._auth_key = auth_key
        logger.add(sys.stderr, level=log_level)

    def run(self):
        threading.Thread(target=lambda: app.run(host='0.0.0.0', port=self._api_port, debug=True, use_reloader=False), daemon=True).start()
        m = QueueManager(address=('0.0.0.0', self._queue_port), authkey=self._auth_key)
        s = m.get_server()
        s.serve_forever()


if __name__ == '__main__':
    smq_server = SmqServer(55555, 44444, b'abr')
    smq_server.run()
